"""
Functions for splitting a notebook file into smaller parts
"""
import string
import re

from soorgeon import exceptions


def find_breaks(nb):
    """Find index breaks based on H2 markdown indexes

    Notes
    -----
    The first element of the returned list may be >0 if the first H2 header
    isn't in the first cell
    """
    breaks = []

    # TODO: this should return named tuples with index and extracted names
    for idx, cell in enumerate(nb.cells):
        # TODO: more robust H2 detector
        if cell.cell_type == 'markdown' and _get_h2_header(cell.source):
            breaks.append(idx)

    if not breaks:
        raise exceptions.InputError('Expected to have at least one markdown '
                                    'with a level 2 header')

    return breaks


def split_with_breaks(cells, breaks):
    """Split a list of cells at given indexes

    Notes
    -----
    Given that the first index has the cell indx of the first H2 header, but
    there may be code in ealier cells, we ignore it. The first split is always
    from 0 to breaks[1]
    """
    breaks = breaks + [None]
    breaks[0] = 0

    cells_split = []

    for left, right in zip(breaks, breaks[1:]):
        cells_split.append(cells[left:right])

    return cells_split


def names_with_breaks(cells, breaks):
    return [_get_h2_header(cells[break_]['source']) for break_ in breaks]


def _sanitize_name(name):
    """Sanitize content of an H2 header to be used as a filename
    """
    # replace all non-aplhanumeric with a dash
    sanitized = re.sub('[^0-9a-zA-Z]+', '-', name.lower())

    # argo does not allow names to start with a digit when using dependencies
    if sanitized[0] in string.digits:
        sanitized = 'section-' + sanitized

    return sanitized


def _get_h2_header(md):
    lines = md.splitlines()

    found = None

    for line in lines:
        match = re.search(r'\s*##\s+(.+)', line)

        if match:
            found = _sanitize_name(match.group(1))

            break

    return found



"""
ProtoTask handles the logic to convert a notebook section into a Ploomber task
"""
from copy import deepcopy
from pathlib import Path

import nbformat
import jupytext
from jinja2 import Template

from soorgeon import io

# TODO: add unit test to ensure parent directories are created
_PICKLING_TEMPLATE = Template("""
{% for product in products %}
Path(product['{{product}}']).parent.mkdir(exist_ok=True, parents=True)
Path(product['{{product}}']).write_bytes(pickle.dumps({{product}}))
{% endfor %}
""")

_UNPICKLING_TEMPLATE = Template("""
{% for up, key in up_and_in %}
{{key}} = pickle.loads(Path(upstream['{{up}}']['{{key}}']).read_bytes())
{% endfor %}
""")


class ProtoTask:
    """A group of cells that will be converted into a Ploomber task
    """

    def __init__(self, name, cells):
        self._name = name
        self._cells = cells

    @property
    def name(self):
        return self._name

    def exposes(self):
        """Return a list of variables that this prototask creates
        """
        pass

    def uses(self):
        """Return a list of variables that this prototask uses
        """
        pass

    def _pickling_cell(self, io):
        """Add cell that pickles the outputs
        """
        _, outputs = io[self.name]

        if outputs:
            pickling = nbformat.v4.new_code_cell(
                source=_PICKLING_TEMPLATE.render(products=outputs))
            pickling.metadata['tags'] = ['soorgeon-pickle']

            return pickling
        else:
            return None

    def _unpickling_cell(self, io, providers):
        """Add cell that unpickles the inputs
        """
        inputs, _ = io[self.name]

        if inputs:
            up_and_in = [(providers.get(input_, self.name), input_)
                         for input_ in inputs]

            unpickling = nbformat.v4.new_code_cell(
                source=_UNPICKLING_TEMPLATE.render(up_and_in=up_and_in))
            unpickling.metadata['tags'] = ['soorgeon-unpickle']

            return unpickling
        else:
            return None

    def _add_parameters_cell(self, cells, upstream):
        """Add parameters cell at the top
        """
        source = ''

        upstream_current = upstream[self.name]

        if upstream_current:
            source += f'upstream = {list(upstream_current)}\n'
        else:
            source += 'upstream = None\n'

        source += 'product = None'

        parameters = nbformat.v4.new_code_cell(source=source)
        parameters.metadata['tags'] = ['parameters']

        return [parameters] + cells

    def _add_imports_cell(self, code_nb, add_pathlib_and_pickle, definitions):
        # FIXME: instatiate this in the constructor so we only build it once
        ip = io.ImportsParser(code_nb)
        source = ip.get_imports_cell_for_task(io.remove_imports(str(self)))

        # FIXME: only add them if they're not already there
        if add_pathlib_and_pickle:
            source = source or ''
            source += '\nfrom pathlib import Path'
            source += '\nimport pickle'

        if definitions:
            names = ', '.join(definitions)
            source = source or ''
            source += f'\nfrom exported import {names}'

        if source:
            cell = nbformat.v4.new_code_cell(source=source)
            cell.metadata['tags'] = ['soorgeon-imports']
            return cell

    def export(self, upstream, io_, providers, code_nb, definitions):
        """Export as a Python string

        Parameters
        ----------
        definitions : dict
            {name: code, ...} mapping with all the function and class
            definitions in the notebook. Used to add an import statement
            to the task
        """

        nb = nbformat.v4.new_notebook()
        # TODO: simplify, make each function return a single cell and then join
        # here

        cells = deepcopy(self._cells)

        # remove import statements from code cells
        # FIXME: remove function definitions and class definitions
        for cell in cells:
            if cell.cell_type == 'code':
                cell['source'] = io.remove_imports(cell['source'])

        # remove empty cells and whitespace-only cells (we may have some after
        # removing imports)
        cells = [cell for cell in cells if cell['source'].strip()]

        cell_unpickling = self._unpickling_cell(io_, providers)

        if cell_unpickling:
            cells = [cell_unpickling] + cells

        cells = self._add_parameters_cell(cells, upstream)

        cell_pickling = self._pickling_cell(io_)

        if cell_pickling:
            cells = cells + [cell_pickling]

        cell_imports = self._add_imports_cell(
            code_nb,
            add_pathlib_and_pickle=cell_pickling or cell_unpickling,
            definitions=definitions)

        pre = [cell_imports] if cell_imports else []

        nb.cells = pre + cells

        # TODO: H2 header should be the top cell

        return jupytext.writes(nb, fmt='py:percent')

    def to_spec(self, io, product_prefix=None):
        """

        Parameters
        ----------
        product_prefix : str
            A prefix to add to all products. If None, it's set to 'output'
        """
        product_prefix = product_prefix or 'output'

        _, outputs = io[self.name]

        # prefix products by name to guarantee they're unique
        products = {
            out: str(Path(product_prefix, f'{self.name}-{out}.pkl'))
            for out in outputs
        }

        # FIXME: check that there isn't an nb key already
        products['nb'] = str(Path(product_prefix, f'{self.name}.ipynb'))

        return {
            'source': str(Path('tasks', self.name + '.py')),
            'product': products
        }

    def __str__(self):
        """Retun the task as string (only code cells)
        """
        return '\n'.join(cell['source'] for cell in self._cells
                         if cell.cell_type == 'code')



import click

from soorgeon import __version__, export


@click.group()
@click.version_option(__version__)
def cli():
    pass


@cli.command()
@click.argument('path', type=click.Path(exists=True))
@click.option('--log', '-l', default=None)
@click.option('--product-prefix',
              '-p',
              default=None,
              help='Prefix for all products')
def refactor(path, log, product_prefix):
    """Refactor a monolithic notebook
    """
    export.from_path(path, log, product_prefix=product_prefix)
    click.secho(f'Finished refactoring {path!r}, use Ploomber to continue.',
                fg='green')
    click.echo("""
Install:
    $ pip install ploomber

List tasks:
    $ ploomber status

Execute pipeline:
    $ ploomber build

Plot pipeline (note: this requires pygraphviz):
    $ ploomber plot

Documentation: https://docs.ploomber.io
""")



"""
CLI for downloading Kaggle notebooks for integration testing
"""
from functools import partial
import zipfile
import shutil
from pathlib import PurePosixPath, Path

import click
import jupytext
import papermill as pm
from kaggle import api


def process_index(index):
    return {PurePosixPath(i['url']).name: _add_partial(i) for i in index}


def _add_partial(d):
    url_data = PurePosixPath(d['data'])
    is_competition = 'c' in url_data.parts
    fn = download_from_competition if is_competition else download_from_dataset

    if is_competition:
        name = url_data.name
    else:
        name = str(PurePosixPath(*url_data.parts[-2:]))

    kwargs = dict(name=name)

    if d.get('files'):
        kwargs['files'] = d['files']

    return {**d, **dict(partial=partial(fn, **kwargs))}


def download_from_competition(name, files=None):
    # FIXME: add support for more than one file
    api.competition_download_cli(name, file_name=files)

    if not files:
        with zipfile.ZipFile(f'{name}.zip', 'r') as file:
            file.extractall('input')
    else:
        Path('input').mkdir()
        shutil.move(files, Path('input', files))


def download_from_dataset(name):
    api.dataset_download_cli(name, unzip=True, path='input')


@click.group()
def cli():
    pass


@cli.command()
@click.argument('kernel_path')
def notebook(kernel_path):
    click.echo('Downloading notebook...')
    name = PurePosixPath(kernel_path).name
    api.kernels_pull_cli(kernel=kernel_path, path=name)

    click.echo('Converting to .py...')
    ipynb = Path(name, f'{name}.ipynb')
    py = Path(name, 'nb.py')
    nb = jupytext.read(ipynb)
    # TODO: remove cells that are !pip install ...
    jupytext.write(nb, py, fmt='py:percent')
    ipynb.unlink()


# FIXME: have a single command that detects if it's a competition or not
# update CONTRIBUTING.md
# FIXME: add files arg
@cli.command()
@click.argument('name')
def competition(name):
    download_from_competition(name=name)


@cli.command()
@click.argument('name')
def dataset(name):
    download_from_dataset(name=name)


@cli.command()
@click.argument('path', type=click.Path(exists=True))
def test(path):
    nb = jupytext.read(path, fmt='py:percent')
    click.echo('Generating test.ipynb...')
    jupytext.write(nb, 'test.ipynb')
    click.echo('Executing test.ipynb...')
    pm.execute_notebook('test.ipynb', 'test.ipynb')


if __name__ == '__main__':
    cli()



import re
import ast
from glob import glob
from os.path import basename
from os.path import splitext

from setuptools import find_packages
from setuptools import setup

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('src/soorgeon/__init__.py', 'rb') as f:
    VERSION = str(
        ast.literal_eval(
            _version_re.search(f.read().decode('utf-8')).group(1)))

REQUIRES = [
    'jupytext',
    'parso',
    'nbformat',
    'jinja2',
    'pyyaml',
    'click',
    'isort',
]

DEV = [
    'pkgmt',
    'pytest',
    'yapf',
    'flake8',
    'invoke',
    'twine',
    'ipython',
    'ploomber',
    # to download data for running _kaggle notebooks
    'kaggle',
    # to run some of the examples
    'pandas',
    'scikit-learn',
    'seaborn',
    'papermill',
    'pkgmt',
    'twine',
]

DESCRIPTION = ('Convert monolithic Jupyter notebooks'
               ' into maintainable pipelines.')

setup(
    name='soorgeon',
    version=VERSION,
    description=DESCRIPTION,
    license=None,
    author='Eduardo Blancas',
    author_email='hello@ploomber.io',
    url='https://github.com/ploomber/soorgeon',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    package_data={"": []},
    classifiers=[],
    keywords=[],
    install_requires=REQUIRES,
    extras_require={
        'dev': DEV,
    },
    entry_points={
        'console_scripts': ['soorgeon=soorgeon.cli:cli'],
    },
)
